//
//------------------------------------------------------------------------------
// Name    : TM1637.bas
// Author  : Jerry Messina
// Date    : 27 JAN 2023
// Version : 1.2
// Notes   : TM1637 display driver
//         : adapted from arduino code https://github.com/avishorp/TM1637
//         : ref - https://www.makerguides.com/tm1637-arduino-tutorial
//         : hdw - DILTAO TM1637 4-digit clock module (ASIN B0B73MBP8V)
//
// ver 1.2
//  - change default timing to 100us (see bit timing below)
//  - add user-callback event to Init()
//  - add DisplayClock()
// ver 1.1
//  - add overload setSegments() for const data array
//  - add #option TM1637_MAX_DIGITS to support up to 6 digit displays
//------------------------------------------------------------------------------
//
module TM1637

// CLK and DIO pins
#option TM1637_CLK = PORTB.0        // pullup optional (see CLK pin mode below)
#option TM1637_DIO = PORTB.1        // requires a pullup

// set CLK pin mode
//   true = open-drain (requires pullup), false = push-pull
// TM1637 does not drive the CLK, so it can be push-pull
#option TM1637_OD_CLK = true

// bit timing, in usec
// note: the app circuit in the TM1637 datasheet shows 10K pullup/100pF to GND on the CLK and DIO lines,
// which would give a risetime of ~1us on the low to high signal transitions
// however, the DILTAO module had a measured rise time of ~100us in OD mode, so more like 0.01uF!
// setting the CLK to push-pull mode (TM1637_OD_CLK=false) allows you to reduce the time
// to about 50us, but with the caps removed the module was able to handle a bit_delay=1us 
// with TM1637_OD_CLK = true
#option TM1637_BIT_DELAY = 100        // 200us total clk period (5KHz)
#if (TM1637_BIT_DELAY = 0)
  #warning "TM1637 should have a min bit_delay of 1us"
#endif
const bit_delay_us = TM1637_BIT_DELAY

// number of display digits
#option TM1637_MAX_DIGITS = 4
#if (TM1637_MAX_DIGITS > 6)
  #error TM1637_MAX_DIGITS, "TM1637 supports a max of 6 digits"
#endif
public const MAX_DIGITS = TM1637_MAX_DIGITS

// ack code
// DIO is allowed to change while CLK is low, but must remain stable while CLK is high to
// avoid generating a false start/stop condition (DIO is clocked on the rising edge of CLK)
// the original arduino code asserts DIO low during the ACK phase once
// it reads the DIO low from the TM1637 (perhaps to try and prevent glitches on the DIO line??)
// this should not be required, but option TM1637_ARDUINO_ACK enables/disables that
// true=use original arduino method(assert DIO low), false=skip asserting DIO low
#option TM1637_ARDUINO_ACK = true

// pin definitions (from options)
dim CLK as TM1637_CLK.TM1637_CLK@,
    DIO as TM1637_DIO.TM1637_DIO@

// TM1637 commands
const
    TM1637_DATA_CMD = $40,      // Data command setting (auto incr address mode)
    TM1637_ADDR_CMD = $C0,      // Address command setting (C0-C5)
    TM1637_CTRL_CMD = $80,      // Display and control command setting
    TM1637_READ_KEY = $42       // Read key scan data                  

//
// DILTAO module LED segment layout (four digit clock 88:88)
//
//      A
//     ---
//  F |   | B
//     -G-
//  E |   | C
//     ---
//      D
public const
    SEG_A   = %00000001,
    SEG_B   = %00000010,
    SEG_C   = %00000100,
    SEG_D   = %00001000,
    SEG_E   = %00010000,
    SEG_F   = %00100000,
    SEG_G   = %01000000,
    SEG_DP  = %10000000     // clock module ':' colon after digit2

public const digitToSegment() as byte = 
(
 // xGFEDCBA
   %00111111,    // 0
   %00000110,    // 1
   %01011011,    // 2
   %01001111,    // 3
   %01100110,    // 4
   %01101101,    // 5
   %01111101,    // 6
   %00000111,    // 7
   %01111111,    // 8
   %01101111,    // 9
   %01110111,    // A
   %01111100,    // b
   %00111001,    // C
   %01011110,    // d
   %01111001,    // E
   %01110001     // F
)

public const minusSegment as byte = SEG_G


// pin IO controls
// assumes open-drain mode with pullups on CLK and DIO pins
sub bit_delay()
    delayus(bit_delay_us)
end sub

inline sub CLK_HIGH()
  #if (TM1637_OD_CLK)   // open-drain mode
    input(CLK)
  #else                 // push-pull
    high(CLK)
  #endif    
end sub

inline sub CLK_LOW()
    low(CLK)
end sub

inline sub DIO_HIGH()
    input(DIO)
end sub

inline sub DIO_LOW()
    low(DIO)
end sub

// init IO and call optional user event
// the event can be used for any addtl setup (ie enable wpu, etc)
type event_t = event()

public sub init(ioinit as event_t=0)
    // set the pin direction and default value
    // both pins are set as inputs, allowing the pull-up resistors to pull them up
    CLK_HIGH()
    DIO_HIGH()
    
    // call user event to finish setup (optional)
    ioinit()
end sub

// generate a START (DIO high to low while CLK high)
public sub start()
    CLK_HIGH()
    bit_delay()
    DIO_LOW()
    bit_delay()
end sub

// generate a STOP (DIO low to high while CLK high)
public sub stop()
    DIO_LOW()
    bit_delay()
    CLK_HIGH()
    bit_delay()
    DIO_HIGH()
    bit_delay()
end sub

// write byte and get display ack response
// returns state of ack... 0=ACK, 1=no ACK/no response 
public function writebyte(b as byte) as bit
    dim i as byte
    
    // clock out 8 data bits, lsb to msb
    for i = 0 to 7
        // CLK low
        CLK_LOW()
        bit_delay()

        // set data bit
        if (b.bits(0) = 1) then
            DIO_HIGH()
        else
            DIO_LOW()
        endif
        bit_delay()
        
        // CLK high
        CLK_HIGH()
        bit_delay()

        // next data bit
        b = b >> 1
    next

    // get acknowledge (9th clock)
    // CLK low
    CLK_LOW()
    DIO_HIGH()  // sets pin to input (float pin)
    bit_delay() // allow time for pullups

    // CLK high
    CLK_HIGH()
    bit_delay()
    // read ack from display
    result = DIO
  #if (TM1637_ARDUINO_ACK)
    if (result = 0) then
        DIO_LOW()       // set pin to match TM1637 ack output
    endif
  #endif
    CLK_LOW()
    bit_delay()
end function

// set display brightness and onoff state
//  brightness = 0 (min) to 7 (max)
//  onoff = 0 (off), 1 (on)
public sub displayCmd(brightness as byte = 7, onoff as bit = 1)
    brightness = brightness and $07         // 0=min, 7=max
    if (onoff = 1) then
        brightness.bits(3) = 1              // display on
    endif        

    // Write TM1637_CTRL_CMD + brightness
    start()
    writebyte(TM1637_CTRL_CMD + brightness)
    stop()
end sub

// write segment data array to display
public sub setSegments(byref segments() as byte, pos as byte=0)
    dim i as byte
    
    // Write TM1637_DATA_CMD
    start()
    writebyte(TM1637_DATA_CMD)
    stop()

    // Write TM1637_ADDR_CMD + display register address (C0-C5)
    if (pos > 5) then
        pos = 5             // note: the display will ignore any addr > 5
    endif        
    start()
    writebyte(TM1637_ADDR_CMD or pos)
    // Write the data bytes
    for i = 0 to (MAX_DIGITS-1)
        writebyte(segments(i))
    next
    stop()
end sub

// write const segment data array to display (overload)
public sub setSegments(byrefconst segments() as byte, pos as byte=0)
    dim data(MAX_DIGITS) as byte
    
    // copy const data to a ram array and write 
    data = segments
    setSegments(data, pos)
end sub

// clear display
public sub clearDisplay()
    dim data(MAX_DIGITS) as byte
    
    clear(data)
    setSegments(data)
end sub

// display a clock "DD:DD"
// four-digit numeric string w/optional colon
public sub DisplayClock(s as string)
    dim data(MAX_DIGITS) as byte

    clear(data)
    FSR0 = addressof(data)
    FSR1 = addressof(s)
    while (INDF1 <> 0)
        select (INDF1)
            case byte("0") to byte("9"):
                INDF0 = digitToSegment(INDF1 - $30)
            case byte(" "):
                INDF0 = 0
            case byte("-"):
                INDF0 = minusSegment
        end select
        if (INDF1 = byte(":")) then   // special case ':'
            data(1) = data(1) or SEG_DP
        else
            WREG = POSTINC0
        endif
        WREG = POSTINC1
    end while
    setSegments(data)
end sub

// read key scan data
// result byte is:  x x x K2 K1 S2 S1 S0
public function readKey() as byte
    dim scan, i as byte
    
    start()

    writebyte(TM1637_READ_KEY)  // read key scan data     
    DIO_HIGH()                  // set DIO to input mode
    scan = 0
    for i = 0 to 7              // start reading from the low
        CLK_LOW()               // clk = 0
        bit_delay()
        CLK_HIGH()              // clk = 1
        bit_delay()
        scan = scan >> 1
        if (DIO = 1) then
            scan.bits(7) = 1
        endif
    next

    // get acknowledge (9th clock)
    // CLK low
    CLK_LOW()
    DIO_HIGH()  // sets pin to input (float pin)
    bit_delay()

    // CLK to high
    CLK_HIGH()
    bit_delay()
  #if (TM1637_ARDUINO_ACK)
    // read ack from display
    if (DIO = 0) then
        DIO_LOW()       // set pin to match TM1637 ack output
    endif
  #endif
    CLK_LOW()
    bit_delay()

    stop()
    result = scan
end function

//
// module init
//
init()

end module


